from functools import wraps
from typing import Any, Callable, Optional, TypeVar, Union, cast, overload

from .timer import Timer

# TODO: Use ParamSpec et al when only python version >=3.10 is supported
F = TypeVar("F", bound=Callable[..., Any])


@overload
def timed(__fn: F) -> F:
    ...


@overload
def timed(*, name: Optional[str] = None, **timer_kwargs: Any) -> Callable[[F], F]:
    ...


def timed(
    __fn: Optional[F] = None, *, name: Optional[str] = None, **timer_kwargs: Any
) -> Union[F, Callable[[F], F]]:
    """
    Function decorator. Decorated functions are wrapped by
    a timer named by the function's module and qualified name.
    (unless explicitly named)
    Any arguments that could be passed to a klokke.Timer constructor
    can be passed as keyword arguments to the decorator instead
    """

    def _timed(fn: F) -> F:
        _name = name or f"{fn.__module__}.{fn.__qualname__}"

        @wraps(fn)
        def wrapper(*args: Any, **kwargs: Any) -> Any:
            with Timer(_name, **timer_kwargs):
                return fn(*args, **kwargs)

        return cast(F, wrapper)

    if __fn is not None:
        return _timed(__fn)

    return _timed



import time
from collections import defaultdict
from copy import deepcopy
from dataclasses import dataclass, field
from typing import Any, Callable, DefaultDict, List, Optional


def _pad_newlines(s: str) -> str:
    return s.replace("\n", "\n  ")


class TimerError(Exception):
    pass


class TimerContext:
    _global_instance = None

    def __init__(self) -> None:
        """
        Creates a TimerContext instance.
        (NB: you generally do not want to instantiate this yourself)
        """
        self._stack: List["Timer"] = []

    def add_to_stack(self, timer: "Timer") -> None:
        self._stack.append(timer)

    def pop_stack(self) -> "Timer":
        return self._stack.pop()

    def current_timer(self) -> Optional["Timer"]:
        if self._stack:
            return self._stack[-1]
        return None

    @classmethod
    def get(cls) -> "TimerContext":
        if cls._global_instance is None:
            cls._global_instance = TimerContext()
            cls._global_instance.add_to_stack(Timer("__global_timer__"))
        return cls._global_instance


@dataclass
class Elapsed:
    n_executed: int = field(default=0)
    total_time: float = field(default=0.0)
    sub_timers: DefaultDict[str, "Elapsed"] = field(
        default_factory=lambda: defaultdict(Elapsed)
    )

    def __add__(self, other: "Elapsed") -> "Elapsed":
        merged_sub_timers = deepcopy(self.sub_timers)
        for k, v in other.sub_timers.items():
            merged_sub_timers[k] += v
        return Elapsed(
            n_executed=self.n_executed + other.n_executed,
            total_time=self.total_time + other.total_time,
            sub_timers=merged_sub_timers,
        )

    def __str__(self) -> str:
        s = f"{self.total_time} seconds"
        if self.sub_timers:
            s += "\nOf which:\n  " + "\n  ".join(
                f"{n}: {_pad_newlines(str(t))}" for n, t in self.sub_timers.items()
            )
        return s


@dataclass
class Timer:
    name: str
    elapsed: Elapsed = field(init=False, default_factory=Elapsed)
    start: float = field(init=False, default=0.0)
    end: float = field(init=False, default=0.0)
    # TODO: Make autoprint be keyword only when only supporting python > 3.10
    autoprint: Optional[Callable[[str], Any]] = field(default=None)

    def __str__(self) -> str:
        s = f"{self.name}: {self.elapsed}"
        return s

    def __enter__(self) -> "Timer":
        context = TimerContext.get()
        context.add_to_stack(self)
        self.start = time.time()
        self.elapsed.n_executed += 1
        return self

    def __exit__(self, *_: Any, **__: Any) -> None:
        self.end = time.time()
        self.elapsed.total_time += self.end - self.start
        context = TimerContext.get()
        # Check that current timer context is ourselves,
        # otherwise something terrible is happening
        current_timer_stack = context.pop_stack()
        if current_timer_stack is not self:
            raise TimerError(
                f"Popping context stack on exiting timer with id {id(self)}"
                f" resulted in popping timer of id {id(current_timer_stack)}"
            )
        parent_timer = context.current_timer()
        if parent_timer is not None:
            parent_timer.track_sub_timer(self)

        if self.autoprint is not None:
            self.autoprint(str(self))

    def track_sub_timer(self, sub_timer: "Timer") -> None:
        self.elapsed.sub_timers[sub_timer.name] += sub_timer.elapsed



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['klokke']

package_data = \
{'': ['*']}

setup_kwargs = {
    'name': 'klokke',
    'version': '0.1.0',
    'description': 'A small library for high level measurement of code execution time',
    'long_description': '# klokke\nklokke is a small library for high level measurement of code execution time. klokke keeps track of running timers, and nests them according to execution order. This allows you to keep track of which sub-sections of code contributes the most to your overall time spent.\nIn addition, a decorator `@timed` is provided for timing function execution.\n\n## Examples\nTiming a snippet of code:\n```python\n>>> from time import sleep\n>>> from klokke import Timer\n>>> with Timer("Something expensive") as timer:\n...     sleep(5)\n... \n>>> print(timer)\nSomething expensive: 5.004069805145264 seconds\n```\n\nNesting timers:\n```python\n>>> from time import sleep\n>>> from klokke import Timer\n>>> with Timer("outer") as outer:\n...     with Timer("inner") as inner:\n...         sleep(1)\n...     sleep(2)\n... \n>>> print(outer)\nouter: 3.0084269046783447 seconds\nOf which:\n  inner: 1.0031757354736328 seconds\n```\n\nUsing a decorator:\n```python\n>>> from time import sleep\n>>> from klokke import Timer, timed\n>>> @timed\n... def foo():\n...     sleep(2)\n...     return 5\n... \n>>> with Timer("bar") as t:\n...     x = foo() + foo()\n... \n>>> print(t)\nbar: 4.010395050048828 seconds\nOf which:\n  __main__.foo: 4.010124683380127 seconds\n```\n\nSetting up timers to automatically print on completion:\n```python\n>>> from time import sleep\n>>> from klokke import Timer\n>>> with Timer("Something expensive", autoprint=print) as timer:\n...     sleep(5)\n... \nSomething expensive: 5.000231981277466 seconds\n```\n\nAll keyword arguments to `Timer` can also be passed through the `@timed` decorator:\n```python\n>>> from time import sleep\n>>> from klokke import timed\n>>> import logging\n>>> logger = logging.getLogger("foo")\n>>> @timed(name="foo_log", autoprint=logger.warning)\n... def foo():\n...     sleep(1)\n... \n>>> foo()\nfoo_log: 1.0042669773101807 seconds\n```\n',
    'author': 'Magnus Heskestad Waage',
    'author_email': 'magnushwaage@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
